/*
** geometry.c
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include "wiregl/include/wireGL.h"
#include "wiregl/include/wiregl_protocol.h"
#include "wiregl/include/wiregl_client.h"
#include "wiregl/include/wiregl_common.h"

#include "wiregl/util/floatlib.c"
#include "length_table.c"
#include "bitfield.c"
#include "geometry.h"
#include <float.h>
#include <math.h>

extern GLcontext *pipe_server_state;

typedef struct
{
	GLfloat *objectBox;
	GLmatrix compositeMatrix;
} TransformationDataT;

typedef struct
{
	GLfloat scale[3];
	GLfloat translate[3];
	unsigned char mantissaPrecision[3];
} CompressionDataT;

typedef struct
{
	GLfloat m00, m01, m02;
	GLfloat m10, m11, m12;
	GLfloat m20, m21, m22;
} DerivativeMatrixT;

static __inline void __collectTransformationData(TransformationDataT *outData)
{
	outData->objectBox = __glbucket_getobjectbox();
	memcpy(&outData->compositeMatrix, __glbucket_getcompositematrix(), sizeof(GLmatrix));
}

/*
 * is   mi0 * xo + mi1 * yo + mi2 * zo + mi3   num
 * -- = ------------------------------------ = ---
 * ws   m30 * xo + m31 * yo + m32 * zo + m33   den
 *
 * d(is / ws)   mij   m3j * num(i)
 * ---------- = --- - ------------
 *    d(jo)     den      den^2
 *
 *
 * outData[i][j] -> i = Screen Coordinate
 *                  j = Object Coordinate
 */

static __inline void __calculateDerivativeMatrix(DerivativeMatrixT *outData, GLfloat *nearPos, TransformationDataT *inData)
{
	GLfloat den;
	GLfloat denSqr;
	GLfloat numXO, numYO, numZO;
	GLmatrix *mat = &inData->compositeMatrix;

	den = mat->m30 * nearPos[0] + mat->m31 * nearPos[1] + mat->m32 * nearPos[2] + mat->m33;
	denSqr = den * den;

	numXO = mat->m00 * nearPos[0] + mat->m01 * nearPos[1] + mat->m02 * nearPos[2] + mat->m03;
	numYO = mat->m10 * nearPos[0] + mat->m11 * nearPos[1] + mat->m12 * nearPos[2] + mat->m13;
	numZO = mat->m20 * nearPos[0] + mat->m21 * nearPos[1] + mat->m22 * nearPos[2] + mat->m23;

	outData->m00 = mat->m00 / den - mat->m30 * numXO / denSqr;
	outData->m01 = mat->m01 / den - mat->m31 * numXO / denSqr;
	outData->m02 = mat->m02 / den - mat->m32 * numXO / denSqr;

	outData->m10 = mat->m10 / den - mat->m30 * numYO / denSqr;
	outData->m11 = mat->m11 / den - mat->m31 * numYO / denSqr;
	outData->m12 = mat->m12 / den - mat->m32 * numYO / denSqr;

	outData->m20 = mat->m20 / den - mat->m30 * numZO / denSqr;
	outData->m21 = mat->m21 / den - mat->m31 * numZO / denSqr;
	outData->m22 = mat->m22 / den - mat->m32 * numZO / denSqr;
}

static __inline GLfloat min2(GLfloat a, GLfloat b)
{
	return(a<b?a:b);
}

static __inline GLfloat min3(GLfloat a, GLfloat b, GLfloat c)
{
	return(min2(min2(a,b),c));
}

static __inline unsigned char __exponentPrecision(unsigned char inSpan)
{
	char outPrecision = 0;
	while(inSpan != 0)
	{
		inSpan >>= 1;
		outPrecision++;
	}

	return(outPrecision);
}

static __inline void __determinePrecision(CompressionDataT *outData, DerivativeMatrixT *inMatrix, TransformationDataT *inData)
{
	GLfloat xoxs, xoys, xozs, xprec;
	GLfloat yoxs, yoys, yozs, yprec;
	GLfloat zoxs, zoys, zozs, zprec;

	char xdiffE = (char)__extractProperExponent(inData->objectBox[1] - inData->objectBox[0]);
	char ydiffE = (char)__extractProperExponent(inData->objectBox[3] - inData->objectBox[2]);
	char zdiffE = (char)__extractProperExponent(inData->objectBox[5] - inData->objectBox[4]);

	GLviewportstate *viewport = &pipe_server_state->viewport;
	int depthBits = pipe_server_state->buffer.depthbits;
	unsigned int depth = 1 << depthBits;

	xoxs = 1 / fabsf(inMatrix->m00) / viewport->w;
	xoys = 1 / fabsf(inMatrix->m10) / viewport->h;
	xozs = 1 / fabsf(inMatrix->m20) / depth;

	yoxs = 1 / fabsf(inMatrix->m01) / viewport->w;
	yoys = 1 / fabsf(inMatrix->m11) / viewport->h;
	yozs = 1 / fabsf(inMatrix->m21) / depth;

	zoxs = 1 / fabsf(inMatrix->m02) / viewport->w;
	zoys = 1 / fabsf(inMatrix->m12) / viewport->w;
	zozs = 1 / fabsf(inMatrix->m22) / depth;

	xprec = min3(xoxs,xoys,xozs);
	yprec = min3(yoxs,yoys,yozs);
	zprec = min3(zoxs,zoys,zozs);

	outData->mantissaPrecision[0] = xdiffE - __extractProperExponent(xprec) + 1;
#if 0
	// these comparisons are always false!
	
	if(outData->mantissaPrecision[0] < 0)
		outData->mantissaPrecision[0] = 0;
	if(outData->mantissaPrecision[2] < 0)
		outData->mantissaPrecision[2] = 0;
	if(outData->mantissaPrecision[1] < 0)
		outData->mantissaPrecision[1] = 0;
#endif

	if(outData->mantissaPrecision[0] > 23)
		outData->mantissaPrecision[0] = 23;
	outData->mantissaPrecision[1] = ydiffE - __extractProperExponent(yprec) + 1;
	if(outData->mantissaPrecision[1] > 23)
		outData->mantissaPrecision[1] = 23;
	outData->mantissaPrecision[2] = zdiffE - __extractProperExponent(zprec) + 1;
	if(outData->mantissaPrecision[2] > 23)
		outData->mantissaPrecision[2] = 23;
}

static __inline void __processCompressionData(CompressionDataT *outData, TransformationDataT *inData)
{
	DerivativeMatrixT dMatrix;
	GLfloat nearPos[3];

	outData->scale[0] = inData->objectBox[1] - inData->objectBox[0];
	outData->scale[1] = inData->objectBox[3] - inData->objectBox[2];
	outData->scale[2] = inData->objectBox[5] - inData->objectBox[4];

	outData->translate[0] = inData->objectBox[0];
	outData->translate[1] = inData->objectBox[2];
	outData->translate[2] = inData->objectBox[4];
	
	nearPos[0] = (inData->objectBox[0] + inData->objectBox[1]) / 2;
	nearPos[1] = (inData->objectBox[2] + inData->objectBox[3]) / 2;
	nearPos[2] = (inData->objectBox[4] + inData->objectBox[5]) / 2;

	__calculateDerivativeMatrix(&dMatrix, nearPos, inData);
	__determinePrecision(outData, &dMatrix, inData);
}

static __inline size_t __compressedSize(CompressionDataT *compressionData)
{
	size_t totalSize = 0;
	int i;

	for(i = 0; i < 3; i++)
		totalSize += compressionData->mantissaPrecision[i];

	return((totalSize + 7) / 8);
}

static __inline void __compressCoord(BitFieldT *bitField, CompressionDataT *compressionData, 
									 GLfloat coord, int coordNum)
{
	unsigned int mantissa;
	
	coord -= compressionData->translate[coordNum];
	if(compressionData->scale[coordNum] != 0.0)
		coord /= compressionData->scale[coordNum];
	mantissa = __extractMantissa(coord);
	mantissa >>= (23 - compressionData->mantissaPrecision[coordNum]);
	AddBitsToBitField(bitField, (unsigned char *)&mantissa, compressionData->mantissaPrecision[coordNum]);
}

static __inline void __compressVertex(int opcode, CompressionDataT *compressionData, unsigned char *targetData,
									  unsigned char *curData, size_t *dataSize, size_t *compressedDataSize)
{
	int i;

	*dataSize = 12;
	*compressedDataSize = __compressedSize(compressionData);
	GLfloat curCoord;
	BitFieldT outVertex;

	InitBitField(&outVertex);
	
	for(i = 0; i < 3; i++)
	{
		curCoord = ((GLfloat*)curData)[0];
		curData+=4;
		__compressCoord(&outVertex, compressionData, curCoord, i);
	}
}

static __inline void __handleGeometryData(CompressionDataT *compressionData, unsigned char *geom_data_start,
										  unsigned char *geom_data_current, unsigned char *geom_data_end,
										  unsigned char *geom_opcode_start, unsigned char *geom_opcode_current)
{
	unsigned char *pCurOpcode = geom_opcode_start;
	unsigned char *pCurData = geom_data_start, *pTargetData = geom_data_start;
	size_t compressedDataSize;
	int dataSize;

	while(pCurOpcode < geom_opcode_current)
	{
		if(*pCurOpcode == WIREGL_VERTEX3F_OPCODE)
		{
			__compressVertex(*pCurOpcode, compressionData, pTargetData, pCurData, (size_t *) &dataSize, &compressedDataSize);
			pTargetData += compressedDataSize;
			pCurData += dataSize;
		}
		else
		{
			dataSize = packet_length_table[*pCurOpcode];
			if(dataSize == -1)
				dataSize = *pCurData;

			memmove(pTargetData, pCurData, dataSize);
			pTargetData += dataSize;
			pCurData += dataSize;
		}
		pCurOpcode++;
	}
}

void __compressGeometryData(unsigned char *geom_data_start, unsigned char *geom_data_current,
								   unsigned char *geom_data_end, unsigned char *geom_opcode_start,
								   unsigned char *geom_opcode_current)
{
	TransformationDataT transData;
	CompressionDataT compressionData;

	__collectTransformationData(&transData);
	__processCompressionData(&compressionData, &transData);
	__handleGeometryData(&compressionData, geom_data_start, geom_data_current, geom_data_end,
		geom_opcode_start, geom_opcode_current);
}
