/*
** init.c
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include <stdlib.h>

#include "wiregl/include/wiregl_util.h"
#include "wiregl/include/wiregl_client.h"
#include "wiregl/include/wiregl_protocol.h"
#include "wiregl/include/wiregl_common.h"

#include "wiregl/include/wiregl_instrument.h"

#include "wiregl/glcontext/glconfig.h"
#include "wiregl/glcontext/glcontext.h"

WireGLPipeServer *pipe_servers = NULL;
int num_pipe_servers;
unsigned char *pipe_server_opcode_start, *pipe_server_opcode_current;
unsigned char *pipe_server_data_start, *pipe_server_data_current, *pipe_server_data_end;
unsigned int pipe_server_buffer_size;
GLcontext *pipe_server_state;

unsigned char *pipe_server_opcode_start_one, *pipe_server_opcode_start_two;
unsigned char *pipe_server_data_start_one, *pipe_server_data_start_two;

extern GLapi wiregl_compresspack_api[];

int __getWindowSize( int *width_return, int *height_return );

void wireGLPrintExitStatistics( void )
{
	//printf ("dying!\n");
	wireGLGMDeath( );

	int i;
	if ( print_bytes_per_server )
	{
		wireGLInstrumentPrintPerServerCounter( WIREGL_BYTES_PER_SERVER );
	}
	if ( print_bytes_per_frame_per_server )
	{
		for (i = 0 ; i < num_pipe_servers ; i++)
		{
			wireGLInstrumentPrintPerFrameCounter( WIREGL_BYTES_PER_SERVER_PER_FRAME + i );
		}
	}
	if ( print_min_bytes_sent )
	{
		wireGLInstrumentPrintCounter( WIREGL_MIN_BYTES_SENT );
	}
	if ( print_frame_rate )
	{
		wireGLInstrumentPrintPerFrameCounter( WIREGL_FRAME_TIMER );
	}
	if ( print_flush_time )
	{
		wireGLInstrumentPrintPerFrameCounter( WIREGL_FLUSH_TIMER );
	}
	if ( print_send_time )
	{
		wireGLInstrumentPrintPerFrameCounter( WIREGL_SEND_TIMER );
	}
	if ( print_diff_context_time )
	{
		wireGLInstrumentPrintPerFrameCounter( WIREGL_DIFF_CONTEXT_TIMER );
	}
	if ( print_total_bytes_sent )
	{
		wireGLInstrumentPrintCounter( WIREGL_TOTAL_BYTES_SENT );
	}
	if ( print_total_time )
	{
		wireGLInstrumentPrintTimer( WIREGL_TOTAL_TIME );
	}
	wireGLInstrumentShutDown();
}

/* XXX HACK: Which header file does this go? */
extern void __glpack_UpdateClientPointer (int array, int index_size, int *index, int data_size, unsigned char * data);

void wireGLCreateContext()
{
	static int initialized = 0;
	static int atexitinit = 0;
	int buffer_size = 256*1024;
	char *buffer_size_override;
	int i;

	if (!initialized)
	{
		initialized = 1;
		(void) wireGLInitInstrumentation( TO_SCREEN );
		(void) wireGLInitOS();
		(void) wireGLInitWarnings();
	}

	buffer_size_override = getenv( "WIREGL_PIPE_BUFFER_SIZE" );
	if (buffer_size_override)
	{
		sscanf( buffer_size_override, "%d", &buffer_size );
	}

	pipe_server_buffer_size = buffer_size;
	pipe_server_data_start_one = (unsigned char *) wireGLAlloc( buffer_size );
	pipe_server_data_start_two = (unsigned char *) wireGLAlloc( buffer_size );
	pipe_server_opcode_start_one = (unsigned char *) wireGLAlloc( buffer_size );
	pipe_server_opcode_start_two = (unsigned char *) wireGLAlloc( buffer_size );

	pipe_server_opcode_start = pipe_server_opcode_start_one;
	pipe_server_data_start = pipe_server_data_start_one;
	pipe_server_data_current = pipe_server_data_start;
	pipe_server_opcode_current = pipe_server_opcode_start;
	pipe_server_data_end = pipe_server_data_start + buffer_size;

	wireGLClientParseConfig( "wiregl.conf" );

	if (!atexitinit) {
		atexitinit = 1;
		atexit( wireGLPrintExitStatistics );
	}

	for (i = 0 ; i < num_pipe_servers ; i ++)
	{
		pipe_servers[i].connection->wireGLRegisterMemoryForDMA( pipe_servers[i].connection, pipe_server_data_start_one, buffer_size );
		pipe_servers[i].connection->wireGLRegisterMemoryForDMA( pipe_servers[i].connection, pipe_server_data_start_two, buffer_size );
		pipe_servers[i].connection->wireGLRegisterMemoryForDMA( pipe_servers[i].connection, pipe_server_opcode_start_one, buffer_size );
		pipe_servers[i].connection->wireGLRegisterMemoryForDMA( pipe_servers[i].connection, pipe_server_opcode_start_two, buffer_size );
	}

	GLconfig *c = __glconfig_CreateDefaultConfig();

	c->flush = wireGLFlushBuffer;
	c->immediateapi = wiregl_im_api;
	c->packapi = wiregl_pack_api;
	c->updateclientpointer = __glpack_UpdateClientPointer;

	c->vertexCapi = vertexCapi;
	c->vertexBBapi = vertexBBapi;
	c->vertex_count = &vertex_count;
	c->vertex_max = &vertex_max;
	c->bounds_max = &bounds_max;
	c->bounds_min = &bounds_min;

	c->numprojectors = num_pipe_servers;
	for (i = 0 ; i < num_pipe_servers ; i++)
	{
		c->bounds[i].x1 = pipe_servers[i].mural_x;
		c->bounds[i].x2 = pipe_servers[i].mural_x + pipe_servers[i].w;
		c->bounds[i].y1 = pipe_servers[i].mural_y;
		c->bounds[i].y2 = pipe_servers[i].mural_y + pipe_servers[i].h;
	}

	c->viewport.x1 = 0;
	c->viewport.y1 = 0;
	c->viewport.x2 = wiregl_mural_width;
	c->viewport.y2 = wiregl_mural_height;

	c->localviewport.x1 = 0;
	c->localviewport.y1 = 0;
	if (!__getWindowSize(&(c->localviewport.x2), &(c->localviewport.y2))) {
		c->localviewport.x2 = wiregl_mural_width;
		c->localviewport.y2 = wiregl_mural_height;
	}

	__glInit(c);

	for (i=0; i<num_pipe_servers; i++) 
		pipe_servers[i].context = __glCreateContext();

	pipe_server_state = __glCreateContext();

	__glconfig_FreeConfig( c );
}	

void wireGLSwapBuffers()
{
	wireGLFlushAll( pipe_server_state, WIREGL_SWAPBUFFERS_OPCODE );
}

